from spinn_front_end_common._version import __version__  # NOQA
from spinn_front_end_common._version import __version_name__  # NOQA
from spinn_front_end_common._version import __version_month__  # NOQA
from spinn_front_end_common._version import __version_year__  # NOQA
__author__ = 'stokesa6'
